# xspin
# simulation options panel

set s_options	""
set v_options	""
set a_options	""
set c_options	""

proc simulation {} {
	global s_typ;	global l_typ
	global fvars;	global vvars
	global gvars;	global lvars
	global avars;	global svars
	global rvars;	global seed
	global stop

	set stop 0

	catch {destroy .s}
	toplevel .s

	wm title .s "Simulation Options"
	wm iconname .s "SIM"

	frame .s.rframe -relief raised -borderwidth 1m

	label .s.rframe.lb \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Amount of Detail in Trace" \
		-relief sunken -borderwidth 1m

	checkbutton .s.rframe.fo -text "Show Final State" \
		-variable fvars \
		-relief flat \
		-command "set fvars 1"
	checkbutton .s.rframe.vo -text "Extra Verbose" \
		-variable vvars \
		-relief flat 
	checkbutton .s.rframe.ao -text "Process States" \
		-variable avars \
		-relief flat 
	checkbutton .s.rframe.go -text "Global Variables" \
		-variable gvars \
		-relief flat \
		-command { set cyc_typ 0 }
	checkbutton .s.rframe.lo -text "Local Variables" \
		-variable lvars \
		-relief flat 
	checkbutton .s.rframe.so -text "Send Statements" \
		-variable svars \
		-relief flat 
	checkbutton .s.rframe.ro -text "Receive Statements" \
		-variable rvars \
		-relief flat 

	pack append .s.rframe .s.rframe.lb {top pady 4 frame w fillx} \
		.s.rframe.fo {top pady 4 frame w} \
		.s.rframe.ao {top pady 4 frame w} \
		.s.rframe.go {top pady 4 frame w} \
		.s.rframe.lo {top pady 4 frame w} \
		.s.rframe.ro {top pady 4 frame w} \
		.s.rframe.so {top pady 4 frame w} \
		.s.rframe.vo {top pady 4 frame w}
	pack append .s .s.rframe {left}

	frame .s.mesg -relief raised -borderwidth 1m
	label .s.mesg.loss0 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "A Full Queue" \
		-relief sunken -borderwidth 1m
	radiobutton .s.mesg.loss1 -text "Blocks New Msgs" \
		-variable l_typ -value 0 \
		-relief flat
	radiobutton .s.mesg.loss2 -text "Loses New Msgs" \
		-variable l_typ -value 1 \
		-relief flat
	pack append .s.mesg .s.mesg.loss0 {top pady 4 frame w fillx}
	pack append .s.mesg .s.mesg.loss1 {top pady 4 frame w}
	pack append .s.mesg .s.mesg.loss2 {top pady 4 frame w}
	pack append .s .s.mesg {left frame n}

	frame .s.lframe -relief raised -borderwidth 1m
	label .s.lframe.tl \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Simulation Style" \
		-relief sunken -borderwidth 1m
	radiobutton .s.lframe.gs -text "Guided (using pan.trail)" \
		-variable s_typ -value 1 \
		-relief flat
	radiobutton .s.lframe.rs -text "Random (using seed)" \
		-variable s_typ -value 0 \
		-relief flat
	entry .s.lframe.entry -relief sunken -width 8 
	label .s.lframe.label \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Seed Value"

	pack append .s.lframe .s.lframe.tl {top pady 4 frame w fillx} \
		.s.lframe.gs {top pady 4 frame w} \
		.s.lframe.rs {top pady 4 frame w} \
		.s.lframe.label right \
		.s.lframe.entry right
	pack append .s .s.lframe {top frame n}

	# Run & Cancel button

	pack append .s [button .s.exit -text "Cancel" \
		-command "getseed; stopsim"] {right frame se}
	pack append .s [button .s.run -text "Run" \
		-command "getseed; runsim" ] {right frame se}
}

proc stopsim {} {
	global stop

	set stop 1
	add_log "<stop simulation>"
	catch {destroy .s}
}

proc bld_s_options {} {
	global fvars;	global vvars
	global gvars;	global lvars
	global avars;	global svars
	global rvars;	global l_typ
	global s_typ;	global seed
	global s_options;

	set s_options ""
	if $vvars { set s_options [format "%s -v" $s_options] }
	if $gvars { set s_options [format "%s -g" $s_options] }
	if $lvars { set s_options [format "%s -l" $s_options] }
	if $avars { set s_options [format "%s -p" $s_options] }
	if $svars { set s_options [format "%s -s" $s_options] }
	if $rvars { set s_options [format "%s -r" $s_options] }
	if $l_typ { set s_options [format "%s -m" $s_options] }
	if $s_typ then {
		set s_options [format "%s -t" $s_options]
	} else {
		if "[string length $seed] > 0" {
			set s_options [format "%s -n%s" $s_options $seed]
		}
	}
}

proc getseed {} {
	global seed
	set seed [.s.lframe.entry get]
}

proc runsim {} {
	global s_options
	global s_typ
	global stop

	set stop 0
	bld_s_options
	add_log "<starting simulation>\nspin $s_options"; update
	set s_options [format "|spin %s pan.in" $s_options]

	if "$s_typ == 0" {
		set fd [open pan.in w];		# file not rewritten on -t option
		puts $fd "[.inp.t get 0.0 end]" nonewline
		flush $fd
		catch "close $fd"
	}
	set sbox [mkbox "Simulator Output" "SimOut" "sim.out"]

	set fd [open "$s_options" w+]
	while {[gets $fd line] > -1} {
		catch { .c$sbox.t insert end "$line\n" }
		catch { .c$sbox.t yview -pickplace end }
		update
		if $stop break
	}
	catch {.c$sbox.t yview 0}
	catch "close $fd"
#	catch { destroy .s }
	add_log "<done>"; update
	update
}
