# xspin
# source text references

set boxnr 0
set values	""
set pnames	""

proc savebox {b} {
	set fname [.c$b.f.e1 get]
	if [file_ok $fname]==0 return
	set fd [open $fname w]
	add_log "<saved output in $fname>"
	puts $fd "[.c$b.t get 0.0 end]" nonewline
	catch "flush $fd"
	catch "close $fd"
}

proc mkbox {n m p} {
	global boxnr
	global stop

	incr boxnr
	toplevel .c$boxnr

	text .c$boxnr.t -relief raised -wrap word
	button .c$boxnr.b -text "Close" -command "set stop 1; destroy .c$boxnr"
	pack append .c$boxnr \
		.c$boxnr.t {top expand fill} \
		.c$boxnr.b {right padx 5} \

	if [string length $p]>0 {
		button .c$boxnr.s -text "Stop" -command "set stop 1"
		frame  .c$boxnr.f -relief sunken
		button .c$boxnr.f.e0 -text "Save in: " -command { savebox $boxnr }
		entry  .c$boxnr.f.e1 -relief flat -width 10
		.c$boxnr.f.e1 insert end $p
		pack append .c$boxnr.f \
			.c$boxnr.f.e0 {left padx 5} \
			.c$boxnr.f.e1 {left}
		pack append .c$boxnr \
			.c$boxnr.s {right padx 5} \
			.c$boxnr.f {right padx 5}
	}

	wm title .c$boxnr $n
	wm iconname .c$boxnr $m
	return $boxnr
}

proc hide_state {} {
	catch { destroy .p }
	.menu.table configure -text "Show States" -command "show_state"
}

proc show_state {} {
	global pnames
	global values

	catch {destroy .p}
	toplevel .p

	wm title .p "States"
	wm iconname .p "States"

	scrollbar .p.scroll -command ".p.list yview"
	listbox .p.list -yscroll ".p.scroll set" -relief raised -geometry 20x20

	peruse
	if [string length $pnames]==0 {
		destroy .p
		return
	}
	.menu.table configure -text "Hide States" -command "hide_state"

	while {[string length $pnames] > 0} {
		set pto [string last "|" $pnames]
		set vto [string last "|" $values]
		set pco [string range $pnames [expr $pto+1] end]
		set vco [string range $values [expr $vto+1] end]

		while 1 {
			set nco [string last "+" $vco]
			if { $nco < 0 } break
			set i [string range $vco [expr $nco+1] end]
			set vco [string range $vco 0 [expr $nco-1]]
			if {$i<10} {
				.p.list insert end "$pco @ state  $i"
			} else {
				.p.list insert end "$pco @ state $i"
			}
		}
		set pnames [string range $pnames 0 [expr $pto-1]]
		set values [string range $values 0 [expr $vto-1]]
	}
	pack append .p \
		.p.scroll {right filly} \
		.p.list {left expand fill}

	bind .p.list <Double-Button-1> {
		gettable [selection get]
		}
}

proc peruse {} {
	global values
	global pnames

	if ![file executable pan] {
		show_log
		add_log "no executable pan (do a validation first)"
		return
	}

	set fd [open "|pan -d" w+]
	while {[gets $fd line] > -1} {
		set have [string first "proctype" $line]
		if {$have > -1} {
			set have [expr $have+9]
			set more [string trim [string range $line $have end]]
			set pnames [format "%s|%s" $pnames $more]
			set values [format "%s|" $values]
			set lastval -1
		} else {
			set have [string first "state" $line]
			if {$have > -1} {
				set have [expr $have+6]
				set upto [expr [string first "--" $line]-2]
				set newval [string trim [string range $line $have $upto]]
				if {$newval != $lastval} {
					set values [format "%s+%s" $values $newval]
					set lastval $newval
				}
			}
		}
	}
	catch "close $fd"
}

proc src_line {s} {
	scan $s %d tgt_lnr

	.inp.t tag add hilite $tgt_lnr.0 $tgt_lnr.1000
	.inp.t tag configure hilite -background black -foreground white
	.inp.t yview -pickplace [expr $tgt_lnr-1]
}

proc gettable { w } {
	global boxnr

	if ![file executable pan] {
		destroy .p
		show_log
		add_log "no executable pan command"
		return
	}

	set w [string trim $w "\{\}"]
	set want  [string first "@" $w]
	set pname [string trim [string range $w 0 [expr $want-1]]]
	set sname [string trim [string range $w [expr $want+1] end]]

	mkbox "Proctype $w" "States" ""

	set want -1
	set fd [open "|pan -d" w+]

	# first remove all highlights in the input window
	.inp.t tag remove hilite 0.0 end

	set maxlen 0
	set ln_cnt 0
	while {[gets $fd line] > -1} {
		if {[string first "proctype" $line] > -1} {
			set want [string first $pname $line]
			set has $want
		} else {
			set has [string first "$sname" $line]
			if {$want<0 || $has<0} {
				continue
			}
			set more [string first "line" $line]
			if {$more<0} {
				continue;	# no line-number can't really happen
			}

			set outg [string first "\t$sname" $line]
			if {$outg>-1} {
				# Outgoing transition: highlight it
				scan [string range $line $more end] "line %d =>" where
				src_line $where

				set nn 0; set ss ""; set foo -1
				set foo [string last "state " $line]
				scan [string range $line $foo end] "state %d" nn
				set foo [string first " => " $line]
				set ss  [string range $line [expr $foo+4] end]
				set foo "%s    ==(%s)==>> state %d\n"
				set added [format $foo $sname $ss $nn]
				.c$boxnr.t insert end $added
			} else {
				# Incoming transition

				set nn 0; set ss ""; set foo -1
				scan $line "\tstate %d" nn
				set foo [string first " => " $line]
				set ss  [string range $line [expr $foo+4] end]
				set foo "%s  <<==(%s)==   state %d\n"
				set added [format $foo $sname $ss $nn]
				.c$boxnr.t insert 0.0 $added
			}
			set len [string length $added]
			if $len>$maxlen { set maxlen $len }
			incr ln_cnt
		}
	}
	.c$boxnr.t configure -height $ln_cnt -width $maxlen
	catch "close $fd"
	update
}
