# xspin
# validation options panel

set an_typ 0;	set cyc_typ 0
set ct_typ 0;	set et_typ 1
set nv_typ 1

set e(0)	"Maximum Memory Usage (Mbytes): "
set e(1)	"Estimated State Space Size (states x 10^3): "
set e(2)	"Maximum Search Depth (steps): "

proc validation {} {
	global e;	global ival
	global an_typ;	global nv_typ
	global cyc_typ;	global ct_typ
	global et_typ;	global stop

	set stop 0

	catch {destroy .v}
	toplevel .v

	wm title .v "Validation Options"
	wm iconname .v "VAL"

	frame .v.pan -relief flat

	# Initialize Sliders

	frame .v.memopts -relief flat;				# memory options panel
	frame .v.correct -relief flat;				# correctness properties
	frame .v.oframe  -relief raised -borderwidth 1m;	# error trail options
	frame .v.mesg -relief raised -borderwidth 1m;		# queue loss options
	frame .v.cframe  -relief raised -borderwidth 1m;	# compile time options

	for {set x 0} {$x<=2} {incr x} {
		global e
		global ival

		frame .v.memopts.choice$x -relief flat
		entry .v.memopts.choice$x.e1 -relief sunken -width 10
		label .v.memopts.choice$x.e2 -text $e($x) -relief flat
		.v.memopts.choice$x.e1 insert end $ival($x)

		pack append .v.memopts.choice$x \
			.v.memopts.choice$x.e2 {left  frame w fillx} \
			.v.memopts.choice$x.e1 {right frame e fillx}

		pack append .v.memopts \
			.v.memopts.choice$x { top frame w pady 5 fillx}
	}
	set z .v.correct

	frame $z.rframe  -relief raised -borderwidth 1m

	label $z.rframe.lb \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Correctness Properties" \
		-relief sunken -borderwidth 1m

	radiobutton $z.rframe.sp -text "State Properties" \
		-variable an_typ -value 0 \
		-relief flat \
		-command { set cyc_typ 0 }
	radiobutton $z.rframe.np -text "Non-Progress Cycles" \
		-variable an_typ -value 1 \
		-relief flat 
	radiobutton $z.rframe.ac -text "Acceptance Cycles" \
		-variable an_typ -value 2 \
		-relief flat 

	checkbutton $z.rframe.nv -text "Never Claim (If Present)" \
		-variable nv_typ \
		-relief flat \
		-command { set nv_typ 1 }
	checkbutton $z.rframe.fc -text "Apply Fairness Restriction" \
		-variable cyc_typ \
		-relief flat  \
		-command { if $an_typ==0 then "set cyc_typ 0" }

	pack append $z.rframe \
		$z.rframe.lb {top pady 4 frame w fillx} \
		$z.rframe.nv {top pady 4 frame w} \
		$z.rframe.sp {top pady 4 frame w} \
		$z.rframe.np {top pady 4 frame w} \
		$z.rframe.ac {top pady 4 frame w} \
		$z.rframe.fc {top pady 4 frame w}
	pack append $z $z.rframe {left frame nw filly}

	pack append .v \
		.v.memopts {top frame w} \
		.v.correct {left frame nw filly}

	# Trail options

	label .v.oframe.lb \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Error Trails" \
		-relief sunken -borderwidth 1m

	radiobutton .v.oframe.c0 -text "Don't Stop at Errors" \
		-variable et_typ -value 0 \
		-relief flat 
	radiobutton .v.oframe.c1 -text "Stop at Error and Produce Trail" \
		-variable et_typ -value 1 \
		-relief flat 

	set et_typ 1

	pack append .v.oframe .v.oframe.lb {top pady 4 frame nw fillx} \
		.v.oframe.c1 {top pady 4 frame nw} \
		.v.oframe.c0 {top pady 4 frame nw}
	pack append .v .v.oframe {top frame w expand fill}

	label .v.mesg.loss0 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "A Full Queue" \
		-relief sunken -borderwidth 1m
	radiobutton .v.mesg.loss1 -text "Blocks New Msgs" \
		-variable l_typ -value 0 \
		-relief flat
	radiobutton .v.mesg.loss2 -text "Loses New Msgs" \
		-variable l_typ -value 1 \
		-relief flat
	pack append .v.mesg .v.mesg.loss0 {top pady 4 frame w fillx} \
		.v.mesg.loss1 {top pady 4 frame w} \
		.v.mesg.loss2 {top pady 4 frame w}
	pack append .v .v.mesg {left frame n}

	# Compile time options

	label .v.cframe.lb \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Search Mode" \
		-relief sunken -borderwidth 1m

	radiobutton .v.cframe.ea -text "Exhaustive" \
		-variable ct_typ -value 0 \
		-relief flat 
	radiobutton .v.cframe.sa -text "Supertrace" \
		-variable ct_typ -value 1 \
		-relief flat 

	pack append .v.cframe .v.cframe.lb {top pady 4 frame nw fillx} \
		.v.cframe.ea {top pady 4 frame nw} \
		.v.cframe.sa {top pady 4 frame nw}
	pack append .v .v.cframe {top frame w expand fill}

	# Run & Cancel button

	pack append .v [button .v.exit -text "Cancel" \
		-command "stopval"] {right frame se}
	pack append .v [button .v.run -text "Run" \
		-command "runval" ] {right frame se}
}

proc stopval {} {
	global stop
	global ival

	set ival(0) "[.v.memopts.choice0.e1 get]"
	set ival(1) "[.v.memopts.choice1.e1 get]"
	set ival(2) "[.v.memopts.choice2.e1 get]"

	set stop 1
	add_log "<stop validation>"
	catch {destroy .v}
}

proc log {n} {
	set m 1
	set cnt 0
	while {$m<$n} {
		set m [expr $m*2]
		incr cnt
	}
	return $cnt
}

proc bld_v_options {} {
	global an_typ;	global cyc_typ;
	global et_typ;	global l_typ;
	global ct_typ;	global ival;
	global v_options; global a_options;
	global c_options;

	set a_options "-a"
	if $l_typ==1	{ set a_options [format "%s -m" $a_options] }

	# the Mbytes scale resolution is in millions: 2^20
	set Mbytes [expr 20+[log [.v.memopts.choice0.e1 get]]]

	# the Mstate scale resolution is in thousands: 2^10
	set Mstate [expr 10+[log [.v.memopts.choice1.e1 get]]]
	set Mdepth [.v.memopts.choice2.e1 get]
	set ival(0) "[.v.memopts.choice0.e1 get]"
	set ival(1) "[.v.memopts.choice1.e1 get]"
	set ival(2) "[.v.memopts.choice2.e1 get]"

	if $ct_typ==1	{
		set c_options [format "-DBITSTATE -DMEMCNT=%d" $Mbytes]

		# in supertrace mode: #bits ~~ 128x #states
		# (effectively the #bytes will be ~~ 16x #states)

		set Mstate [expr 7+$Mstate]
		set v_options "-X -m$Mdepth -w$Mstate"
	} else {
		set c_options [format "-DMEMCNT=%d" $Mbytes]

		# in exhaustive mode: #hashtable ~~ #states

		set v_options "-X -m$Mdepth -w$Mstate"
	}
	if $an_typ==1	{ set v_options [format "%s -l" $v_options] }
	if $an_typ==2	{ set v_options [format "%s -a" $v_options] }
	if $cyc_typ==1	{ set v_options [format "%s -f" $v_options] }
	if $et_typ==0	{ set v_options [format "%s -c0" $v_options] }
}

proc runval {} {
	global v_options; global a_options;
	global c_options;
	global stop
	set stop 0

	show_log
	bld_v_options

	set model [.inp.t get 0.0 end]
	set fd [open pan.in w]
	puts $fd $model nonewline
	flush $fd
	catch "close $fd"

	add_log "<starting validation>\nspin $a_options"; update
	catch "exec spin $a_options pan.in" errmsg
		if "$stop || [string length $errmsg]>0" {
			add_log "$errmsg"
			update
			return
		}
	add_log "cc -o pan $c_options pan.c"; update
	catch "exec cc -o pan $c_options pan.c" errmsg
		if "$stop || [string length $errmsg]>0" {
			add_log "$errmsg"
			update
			return
		}

	add_log "pan $v_options"; update
	set fd [open "|pan $v_options" w+]
	set vbox [mkbox "Validator Output" "ValOut" "pan.out"]

	.inp.t tag remove hilite 0.0 end
	while {[gets $fd line] > -1} {
		catch { .c$vbox.t insert end "$line\n" }
		catch { .c$vbox.t yview -pickplace end }

		if {[string first "line" $line]>=0} {
			scan $line "\tline %d" where
			src_line $where
		}

		update
		if $stop break
	}
	catch { .c$vbox.t yview 0 }
	catch "close $fd"
#	catch {destroy .v}
	add_log "<done>"
	update
}
