/***** spin: main.c *****/

/* Copyright (c) 1991 by AT&T Bell Telephone Laboratories, Inc.           */
/* All Rights Reserved. This software is for educational purposes only.   */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.att.com          */

#include <stdio.h>
#include <signal.h>
#include "spin.h"
#include "y.tab.h"

extern int	lineno;
extern FILE	*yyin;
Symbol	*Fname;
int	verbose = 0;
int	analyze = 0;
int	s_trail = 0;
int	m_loss  = 0;
int	nr_errs	= 0;

void
main(argc, argv)
	char *argv[];
{
	Symbol *s;
	int T = (int) time((long *)0);

	signal(SIGPIPE, exit);

	while (argc > 1 && argv[1][0] == '-')
	{	switch (argv[1][1]) {
		case 'a': analyze  =  1; break;
		case 'g': verbose +=  1; break;
		case 'l': verbose +=  2; break;
		case 'm': m_loss   =  1; break;
		case 'n': T = atoi(&argv[1][2]); break;
		case 'p': verbose +=  4; break;
		case 'r': verbose +=  8; break;
		case 's': verbose += 16; break;
		case 'v': verbose += 32; break;
		case 't': s_trail  =  1; break;
		case 'V': printf("%s\n", Version); exit(0);
		default : printf("use: spin -[agmlpqrst] [-nN] file\n");
			  printf("\t-a produce an analyzer\n");
			  printf("\t-g print all global variables\n");
			  printf("\t-l print all local variables\n");
			  printf("\t-m lose msgs sent to full queues\n");
			  printf("\t-nN seed for random nr generator\n");
			  printf("\t-p print all statements\n");
			  printf("\t-r print receive events\n");
			  printf("\t-s print send events\n");
			  printf("\t-v verbose, more warnings\n");
			  printf("\t-t follow a simulation trail\n");
			  exit(1);
		}
		argc--, argv++;
	}
	if (argc > 1)
	{	char outfile[17], cmd[64];
		Fname = lookup(argv[1]);
		mktemp(strcpy(outfile, "/tmp/spin.XXXXXX"));
		sprintf(cmd, "/lib/cpp %s > %s", argv[1], outfile);
		if (system(cmd))
		{	unlink(outfile);
			exit(1);
		} else if (!(yyin = fopen(outfile, "r")))
		{	printf("cannot open %s\n", outfile);
			exit(1);
		}
		unlink(outfile);
	} else
		Fname = lookup("<stdin>");
	srand(T);
	s = lookup("_p"); s->type = PREDEF;
	yyparse();
	exit(nr_errs);
}

yywrap()	/* dummy routine */
{
	return 1;
}

yyerror(s1, s2)
	char *s1, *s2;
{
	extern int yychar; extern char yytext[];
	printf("spin: %s line %d: ", Fname->name, lineno);
	if (s2)
		printf(s1, s2);
	else
		printf(s1);
	if (yychar)
	{	printf("	saw '");
		explain(yychar);
		printf("' %s.\n", (yytext)?yytext:"");
	}
	printf("\n"); fflush(stdout);
	nr_errs++;
}

void
fatal(s1, s2)
	char *s1, *s2;
{
	yyerror(s1, s2);
	fflush(stdout);
	exit(1);
}

char *
emalloc(n)
{	char *tmp = malloc(n);

	if (!tmp)
		fatal("not enough memory", (char *)0);
	memset(tmp, 0, n);
	return tmp;
}

Node *
nn(s, v, t, l, r)
	Symbol *s;
	Node *l, *r;
{
	Node *n = (Node *) emalloc(sizeof(Node));
	n->nval = v;
	n->ntyp = t;
	n->nsym = s;
	n->fname = Fname;
	n->lft  = l;
	n->rgt  = r;
	return n;
}

Node *
rem_lab(a, b, c)
	Symbol *a, *c;
	Node *b;
{
	fix_dest(c, a);	/* in case target is a jump */
	return	nn((Symbol *)0, 0, EQ,
		nn(lookup("_p"), 0, 'p', nn(a, 0, '?', b, (Node *)0), (Node *)0),
		nn(c, 0, 'q', nn(a, 0, NAME, (Node *)0, (Node *)0), (Node *)0));
}

Node *
rem_var(a, b, c)
	Symbol *a;
	Node *b, *c;
{
	Node *tmp;
	tmp = nn(a, 0, '?', b, (Node *)0);
	return nn(c->nsym, 0, 'p', tmp, c->lft);
}

void
explain(n)
{
	switch (n) {
	default:	printf("%c (=%d)", n, n); break;
	case RUN:	printf("run"); break;
	case LEN:	printf("len"); break;
	case OF:	printf("of"); break;
	case CONST:	printf("cont"); break;
	case TYPE:	printf("type"); break;
	case ASGN:	printf("asgn"); break;
	case NAME:	printf("name"); break;
	case CLAIM:	printf("claim"); break;
	case STRING:	printf("string"); break;
	case INIT:	printf("init"); break;
	case ASSERT:	printf("assert"); break;
	case GOTO:	printf("goto"); break;
	case BREAK:	printf("break"); break;
	case MTYPE:	printf("mtype"); break;
	case SEP:	printf("double-colon"); break;
	case IF:	printf("if"); break;
	case FI:	printf("fi"); break;
	case DO:	printf("do"); break;
	case OD:	printf("od"); break;
	case ATOMIC:	printf("atomic"); break;
	case SND:	printf("send"); break;
	case RCV:	printf("receive"); break;
	case PRINT:	printf("printf"); break;
	case TIMEOUT:	printf("timeout"); break;
	case PROCTYPE:	printf("proctype"); break;
	case OR:	printf("||"); break;
	case AND:	printf("&&"); break;
	case EQ:	printf("=="); break;
	case NE:	printf("!="); break;
	case GE:	printf(">="); break;
	case LE:	printf("<="); break;
	case LSHIFT:	printf("<<"); break;
	case RSHIFT:	printf(">>"); break;
	case UMIN:	printf("- (unary minus)"); break;
	case NEG:	printf("! (not)"); break;
	}
}
